# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from abc import ABCMeta, abstractmethod
from typing import TYPE_CHECKING, Optional

if TYPE_CHECKING:
    from collections.abc import Sequence

    from rucio.common import types
    from rucio.core.rse import RseData
    from rucio.transfertool.transfertool import TransferStatusReport


class BittorrentDriver(metaclass=ABCMeta):
    external_name = ''
    required_rse_attrs = tuple()

    @classmethod
    @abstractmethod
    def make_driver(cls: "type[BittorrentDriver]", rse: "RseData", logger: 'types.LoggerFunction' = logging.log) -> "Optional[BittorrentDriver]":
        pass

    @abstractmethod
    def listen_addr(self) -> tuple[str, int]:
        pass

    @abstractmethod
    def add_torrent(self, file_name: str, file_content: bytes, download_location: str, seed_mode: bool = False) -> None:
        pass

    @abstractmethod
    def add_peers(self, torrent_id: str, peers: 'Sequence[tuple[str, int]]') -> None:
        pass

    @abstractmethod
    def get_status(self, request_id: str, torrent_id: str) -> "TransferStatusReport":
        pass
