
'''
This file is automatically generated; Do not edit it. :)
'''
VERSION_INFO = {
    'final': True,
    'version': '36.0.0rc2',
    'branch_nick': 'master',
    'revision_id': '02ac51022e1d551cdef23a909ad7e9ca21ff8b01',
    'revno': 13325
}
